#include <iostream>
#include <fstream>
#include "GameWorld.h"

extern ModelManager modelManager;
extern TextureManager textureManager;
extern Config cfg;

GameWorld::GameWorld(void) {
	nCheckpoints = 0;
	nScenery = 0;
	firstPerson = false;
}

GameWorld::~GameWorld(void) {
	clearWorld();
}

void GameWorld::parseModel(Json::Value& modelNode, ModelEntity& entity) {
	Model* model = modelManager.getOBJModel(modelNode.get("model","").asString());

	entity.modelPointer = model;
	entity.x = (float)modelNode.get("x", 0.0).asDouble();
	entity.y = (float)modelNode.get("y", 0.0).asDouble();
	entity.z = (float)modelNode.get("z", 0.0).asDouble();
	/*
	entity.rx = (float)modelNode.get("rx", 0.0).asDouble();
	entity.ry = (float)modelNode.get("ry", 0.0).asDouble();
	entity.rz = (float)modelNode.get("rz", 0.0).asDouble();
	*/
	entity.pitch = (float)modelNode.get("pitch", 0.0).asDouble();
	entity.yaw = (float)modelNode.get("yaw", 0.0).asDouble();
	entity.roll = (float)modelNode.get("roll", 0.0).asDouble();

	entity.sx = (float)modelNode.get("sx", 1.0).asDouble();
	entity.sy = (float)modelNode.get("sy", 1.0).asDouble();
	entity.sz = (float)modelNode.get("sz", 1.0).asDouble();

	scene.addEntity(entity);

}

void GameWorld::parseScenery(Json::Value& sceneryNode) {
	nScenery = sceneryNode.size();
	scenery = new ModelEntity[nScenery];
	
	for (unsigned int i = 0; i < nScenery; i++) {
		scenery[i] = ModelEntity();
		parseModel(sceneryNode[i], scenery[i]);
	}
}

void GameWorld::parseCheckpoints(Json::Value& checkpNodes) {
	nCheckpoints = checkpNodes.size();
	checkpoints = new ModelEntity[nCheckpoints];
	
	for (unsigned int i = 0; i < nCheckpoints; i++) {
		checkpoints[i] = ModelEntity();
		parseModel(checkpNodes[i], checkpoints[i]);
		checkpoints[i].y = terrain.getHeight(checkpoints[i].x, checkpoints[i].z);

		if (checkpNodes[i].get("isgoal", false).asBool()) {
			player.curCheckp = i;
			player.setLocation(checkpoints[i]);
			player.setRotation(checkpoints[i]);
		}
	}
}

void GameWorld::clearWorld() {
	if (nCheckpoints) {
		delete[] checkpoints;
		nCheckpoints = 0;
	}

	if (nScenery) {
		delete[] scenery;
		nScenery = 0;
	}

	for (std::list<Unit*>::iterator it = units.begin(); it != units.end(); it++) {
		delete *it;
	}
	units.clear();

	scene = Scene();
}

void GameWorld::reloadWorld() {
	loadWorld(curWorldFile);
}

void GameWorld::loadWorld(const std::string& worldFile) {
	clearWorld();

	data.state = playing;
	curWorldFile = worldFile;

	Json::Value root;
	Json::Reader reader;
    std::string worldPath = MAP_PATH + worldFile;

	root = readJSONDoc(worldPath, reader);
	std::string skyboxName = root.get("skybox", "").asString();
	scene.skybox.loadSkybox(skyboxName);

	Json::Value terrainNode = root.get("terrain", "");
	std::string hMap = terrainNode.get("heightMap", "").asString();
	std::string terrainTex = terrainNode.get("texture", "").asString();
	float terrainScale = (float)terrainNode.get("scale", "1.0").asDouble();
	float minH = (float)terrainNode.get("minHeight", "0").asDouble();
	float maxH = (float)terrainNode.get("maxHeight", "100").asDouble();
	int detail = terrainNode.get("detail", "1").asInt();
	
	terrain = Terrain();
	terrain.loadFromPNG(hMap, terrainTex, minH, maxH, terrainScale, detail);
	scene.addEntity(terrain);

	player = Player();
	std::string playerFile = root.get("player", "").asString();
	player.loadPlayer(playerFile);
	player.laps = root.get("laps", 0).asInt();
	scene.addEntity(player);

    Json::Value checkpNode = root.get("checkpoints", "");
	parseCheckpoints(checkpNode);

    Json::Value sceneryNode = root.get("scenery", "");
	parseScenery(sceneryNode);

	scene.camera.y = player.y;
	scene.camera.thirdPerson(player, 10.0f);

	control.loadControls(cfg.controls);

	//sndHit = soundManager.loadSound("hit.wav");
	//sndCheckpoint = soundManager.loadSound("checkpoint.wav");

}

void GameWorld::pollEvent(SDL_Event* e) {
	if ( e->type == SDL_QUIT || 
		(e->type == SDL_KEYDOWN 
			&& e->key.keysym.sym == SDLK_ESCAPE)) {
		running = false;
	}

	control.pollEvent(e);
}

void GameWorld::processInput(float dt) {
	if (control.isDown(forward)) player.accelerate(1.0f);
	else if (control.isDown(backward)) player.brake(1.0f);
	else player.stopAccel();

	if (control.isDown(turnLeft)) player.turn(dt);
	if (control.isDown(turnRight)) player.turn(-dt);
	if (control.isPressed(changeCamera)) {
		firstPerson = !firstPerson;
	}

	if (control.isPressed(dropUnit)) {
		spawnUnit(player);
	}

	if (control.isPressed(reload)) reloadWorld();
}

void GameWorld::processLogic(float dt) {
	

	for (std::list<Unit*>::iterator it = units.begin();
		it != units.end(); it++) {
		(**it).renderLogic(this, dt);
	}
	
	player.renderLogic(this, dt);
	if ((player.hp <= 0.0f) && data.state == playing) {
		data.state = gameover;	
	};

	if (data.state == playing) {
		if (player.collide(checkpoints[player.curCheckp % nCheckpoints])) {
			if (player.curCheckp % nCheckpoints == 0) {
				player.laps--;
			}
			if (player.laps == 0) data.state = win;
			player.curCheckp = (player.curCheckp + 1) % nCheckpoints; 
			data.checkpoint = true;
			//printf("%d laps left\n", player.laps);
		}
	}

}

void GameWorld::renderScene(void) {
	//glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	glViewport(0, 0, cfg.screenW, cfg.screenH);
	gluPerspective((45.0f + (abs(player.velocity / player.maxSpeed) * 45.0f)), (GLfloat)cfg.screenW / (GLfloat)cfg.screenH, 0.1f, 1000.0f);
	//gluPerspective((45.0f), (GLfloat)screenW / (GLfloat)screenH, 0.1f, 1000.0f);
	

	if (firstPerson) {
		scene.camera.firstPerson(player);
	} else {
		scene.camera.thirdPerson(player, 10.0f);
		scene.camera.y = player.y + 2.0f;
		float ch = terrain.getHeight(scene.camera.x, scene.camera.z);
		if (scene.camera.y < ch) {
			scene.camera.y = ch;
		}
	}
	
	scene.setScale(1.0f, 1.0f, 1.0f);
	scene.render();
	
	// Render mirror view
	const GLfloat mirrorW = cfg.screenW / 4.0f;
	const GLfloat mirrorH = mirrorW / 4.0f;
	const GLfloat mirrorX = (cfg.screenW - mirrorW) / 2.0f;
	const GLfloat mirrorY = cfg.screenH - mirrorH;

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glViewport((GLint)mirrorX, (GLint)mirrorY, (GLint)mirrorW, (GLint)mirrorH);
	gluPerspective(45.0f, mirrorW / mirrorH, 0.1f,1000.0f);
	
	scene.camera.freeCamera();
	scene.camera.setLocation(player);
	scene.camera.setRotation(player);
	scene.setScale(-1.0f, 1.0f, -1.0f);
	//scene.render();

	data.speed = fabsf(player.velocity / player.maxSpeed);
	data.hit = player.hit;
	data.hp = player.hp;
	glViewport(0, 0, cfg.screenW, cfg.screenH);
	hud.render(data);

	SDL_GL_SwapBuffers();
}

void GameWorld::spawnUnit(Unit& spawner) {
	Unit* inf = new Unit();
	Model* infModel = modelManager.getOBJModel("sun.obj");
	inf->modelPointer = infModel;

	inf->setLocation(spawner);
	inf->setRotation(spawner);
	inf->setSpeed(fabsf(spawner.maxSpeed) * 2.0f);

	scene.addEntity(*inf);
	units.push_back(inf);
}
