#include "NetSocket.h"
#include "stdafx.h"

#if PLATFORM == PLATFORM_WINDOWS
#include <winsock2.h>
#elif PLATFORM == PLATFORM_MAC || PLATFORM == PLATFORM_UNIX
#include <sys/socket.h>
#include <netinet/in.h>
#include <fcntl.h>

#endif

bool net_init(void) {
#if PLATFORM == PLATFORM_WINDOWS
	WSADATA WsaData;
	return WSAStartup(MAKEWORD(2,2), &WsaData) == NO_ERROR;
#else
	return true;
#endif
}

void net_cleanup(void) {
#if PLATFORM == PLATFORM_WINDOWS
	WSACleanup();
#endif
}

NetAddress::NetAddress(void) {
	port = address = 0;
}

NetAddress::NetAddress(unsigned char a, unsigned char b, unsigned char c, unsigned char d, unsigned short port) {
	this->address = ( a << 24 ) | ( b << 16 ) | ( c << 8 ) | d;
	this->port = port;
}

NetAddress::NetAddress(unsigned int address, unsigned short port) {
	this->address = address;
	this->port = port;
}

unsigned int NetAddress::getAddress(void) const {
	return this->address;
}

unsigned char NetAddress::getA(void) const {
	return (unsigned char)(address >> 24);
}
	
unsigned char NetAddress::getB(void) const {
	return (unsigned char)(address >> 16);
}
	
unsigned char NetAddress::getC(void) const {
	return (unsigned char)(address >> 8);
}
	
unsigned char NetAddress::getD(void) const {
	return (unsigned char)(address);
}
	
unsigned short NetAddress::getPort(void) const { 
	return port;
}
	
bool NetAddress::operator == (const NetAddress& other) const {
	return address == other.address && port == other.port;
}
	
bool NetAddress::operator != (const NetAddress & other) const {
	return ! ( *this == other );
}

NetSocket::NetSocket(void) {
	handler = 0;
}

NetSocket::~NetSocket(void) {
	close();
}

bool NetSocket::open(void) {
	return open(0);
}

bool NetSocket::open(unsigned short port) {
	//assert( !isOpen() );
		
	// create socket

	handler = socket( AF_INET, SOCK_DGRAM, IPPROTO_UDP );

	if ( socket <= 0 ) {
		handler = 0;
		return false;
	}

	// bind to port

	sockaddr_in address;
	address.sin_family = AF_INET;
	address.sin_addr.s_addr = INADDR_ANY;
	address.sin_port = htons( (unsigned short) port );
		
	if ( bind( handler, (const sockaddr*) &address, sizeof(sockaddr_in) ) < 0) {
		close();
		return false;
	}


	return true;
}

void NetSocket::close(void) {
	if (socket != 0) {
#if PLATFORM == PLATFORM_WINDOWS
		closesocket(handler);
#else
		::close(handler);
#endif
		handler = 0;
	}
}

bool NetSocket::isOpen(void) const {
	return handler != 0;
}

bool NetSocket::setNonblock(bool nonBlock) {
#if PLATFORM == PLATFORM_MAC || PLATFORM == PLATFORM_UNIX
		
	int nonBlocking = (int)nonBlock;
	if (fcntl( handler, F_SETFL, O_NONBLOCK, nonBlocking ) == -1) {
		close();
		return false;
	}
			
#elif PLATFORM == PLATFORM_WINDOWS
		
	DWORD nonBlocking = (DWORD)nonBlock;
	if (ioctlsocket(handler, FIONBIO, &nonBlocking) != 0) {
		close();
		return false;
	}

#endif

	return true;
}

bool NetSocket::send(const NetAddress& dest, const void* data, int size) {
		
	if (socket == 0)
		return false;
		
	sockaddr_in address;
	address.sin_family = AF_INET;
	address.sin_addr.s_addr = htonl(dest.getAddress());
	address.sin_port = htons((unsigned short)dest.getPort());

	int sent_bytes = sendto(handler, (const char*)data, size, 0, (sockaddr*)&address, sizeof(sockaddr_in));

	return sent_bytes == size;
}

int NetSocket::receive(NetAddress& sender, void* data, int size) {

	if ( socket == 0 )
		return false;
			
	#if PLATFORM == PLATFORM_WINDOWS
	typedef int socklen_t;
	#endif
			
	sockaddr_in from;
	socklen_t fromLength = sizeof( from );

	int received_bytes = recvfrom(handler, (char*)data, size, 0, (sockaddr*)&from, &fromLength );

	if ( received_bytes <= 0 )
		return 0;

	unsigned int address = ntohl( from.sin_addr.s_addr );
	unsigned int port = ntohs( from.sin_port );

	sender = NetAddress( address, port );

	return received_bytes;
}
