#include "Potato.h"
#include "World.h"
#include "IntroWorld.h"
#include "GameWorld.h"
#include "config.h"

#include "ModelManager.h"
#include "TextureManager.h"

#include "NetSocket.h"
#include "NetClientWorld.h"
#include "NetServerWorld.h"

ModelManager modelManager;
TextureManager textureManager;
Config cfg;
SDL_Surface* screen;


void InitGL(int Width, int Height)
{
	glewInit();
	glViewport(0, 0, Width, Height);
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	
	glClearDepth(1.0);
	glDepthFunc(GL_LESS);
	glEnable(GL_DEPTH_TEST);
	glShadeModel(GL_SMOOTH);

	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	//glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);
    glFrontFace(GL_CCW);

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	gluPerspective(45.0f,(GLfloat)Width/(GLfloat)Height,0.1f,500.0f);
	glMatrixMode(GL_MODELVIEW);

	srand((unsigned int)time(NULL));
}

void setupScreen(void) {
	/* Create a OpenGL screen */
	if ((screen = SDL_SetVideoMode(cfg.screenW, cfg.screenH, 0, SDL_OPENGL | (cfg.fullscreen ? SDL_FULLSCREEN : 0))) == NULL ) {
		fprintf(stderr, "Unable to create OpenGL screen: %s\n", SDL_GetError());
		SDL_Quit();
		exit(2);
	}


	SDL_WM_SetCaption("POTATO", NULL);
	InitGL(cfg.screenW, cfg.screenH);
}

World* renderWorld(World& w) {
	bool done = false;
	int t_prev_interp;
	int t_prev_render;

	int t_prev_frames;
	int t_prev_input;

    t_prev_input = t_prev_frames = t_prev_render = t_prev_interp = SDL_GetTicks();
	float dt;

	int frames = 0;
	w.setup();
	while (w.running) {
		
		int t = SDL_GetTicks();
		dt = (float)(t - t_prev_interp) / (float)cfg.gameSpeed;
		if ((t - t_prev_interp) == 0) continue;

		SDL_Event event;
		if (t - t_prev_input > (int)randnum((float)cfg.minInputLatency, (float)cfg.maxInputLatency)) {
			
			if (randnum(0.0f, 100.0f) > (float)cfg.dropProbability)
				while ( SDL_PollEvent(&event) ) {	
					w.pollEvent(&event);
				}

			t_prev_input = t;
		}
		
		w.processInput(dt);
		w.processLogic(dt);

		if (cfg.fps <= 0 || (float)(t - t_prev_render) >= 1000.0f / (float)cfg.fps) {
			w.renderScene();
			t_prev_render = t;
			frames++;
		}

		if (frames >= GAME_SPEED) {
			printf("FPS: %.2f\n", frames / (((float)(t - t_prev_frames) / 1000.0)));
			frames = 0;
			t_prev_frames = t;
		}

		t_prev_interp = t;
	}

	return w.nextWorld;
}

int main(int argc, char **argv) 
{ 
	if ( SDL_Init(SDL_INIT_VIDEO) < 0 ) {
		fprintf(stderr, "Unable to initialize SDL: %s\n", SDL_GetError());
		exit(1);
	}

	cfg.load();

	setupScreen();

	net_init();

	World* curWorld;
	World* nextWorld;

	if (argc >= 2 && argv[1][0] == 's') {
		printf("Server mode.\n");
		curWorld = new NetServerWorld();
		((GameWorld*)curWorld)->loadWorld(cfg.initialMap);

	} else if (argc >= 2 && argv[1][0] == 'c') {
		printf("Client mode.\n");
		curWorld = new NetClientWorld();
		NetAddress addr(127,0,0,1,9666);
		((NetClientWorld*)curWorld)->connectTo(addr);

	} else {
		if (!cfg.skipIntro) {
			curWorld = new IntroWorld(cfg.initialMap);
		} else {
			curWorld = new GameWorld();
			((GameWorld*)curWorld)->loadWorld(cfg.initialMap);
		}
	}

	while (curWorld) {
		nextWorld = renderWorld(*curWorld);
		delete curWorld;
		curWorld = nextWorld;
	}

	net_cleanup();
	SDL_Quit();
	return 0;
}
